﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using XDesigner.ORM.DOM;
using System.Data ;
using XDesigner.Common;
using System.Collections ;

namespace XDesigner.ORM
{
    public class ORMEngine : IDisposable 
    {
        private static System.LocalDataStoreSlot myDataStoreSlot
            = System.Threading.Thread.AllocateNamedDataSlot(Guid.NewGuid().ToString());

        /// <summary>
        /// Get local instance of current thread
        /// </summary>
        public static ORMEngine LocalInstance
        {
            get
            {
                return System.Threading.Thread.GetData(myDataStoreSlot) as ORMEngine;
            }
        }

        private static ORMEngine myGlobalInstance = new ORMEngine();

        /// <summary>
        /// Get global instance of framework
        /// </summary>
        public static ORMEngine MyGlobalInstance
        {
            get { return myGlobalInstance; }
        }

        private static System.LocalDataStoreSlot myContextSlot
            = System.Threading.Thread.AllocateNamedDataSlot(Guid.NewGuid().ToString());
        /// <summary>
        /// Get current ORM context 
        /// </summary>
        public static ORMContext CurrentContext
        {
            get
            {
                return System.Threading.Thread.GetData(myContextSlot) as ORMContext;
            }
        }

        internal static void SetCurrentContext(ORMContext context)
        {
            System.Threading.Thread.SetData(myContextSlot, context);
        }

        public ORMEngine()
        {
            System.Threading.Thread.SetData(myDataStoreSlot, this);
        }

        public ORMEngine(System.Data.IDbConnection connection)
        {
            _DataBase.Connection = connection;
        }

        private string _Name = "StandardORM";

        public string Name
        {
            get { return _Name; }
            set { _Name = value; }
        }

        private ORMProjectInfo _Project = new ORMProjectInfo();
        /// <summary>
        /// ORM-DOM document
        /// </summary>
        public ORMProjectInfo Project
        {
            get { return _Project; }
            set { _Project = value; }
        }

        private System.Collections.Generic.Dictionary<ORMTypeInfo, ORMTypeMappingHelper> 
            myMappingHelpers = new Dictionary<ORMTypeInfo, ORMTypeMappingHelper>();

        public string GenerateHelperSourceCode(Type type)
        {
            if (type == null)
            {
                throw new ArgumentNullException("type");
            }
            ORMTypeInfo ot = this.Project.Types[type];
            if (ot == null)
            {
                throw new NotSupportedException(type.FullName);
            }
            ORMContext context = new ORMContext( this );
            string code = CodeGenerator.GenerateClassCode(context, ot, null);
            return code;
        }

        /// <summary>
        /// Get mapping helper
        /// </summary>
        /// <param name="info"></param>
        /// <returns></returns>
        public ORMTypeMappingHelper GetHelper(ORMTypeInfo info)
        {
            if (myMappingHelpers.ContainsKey(info))
            {
                return myMappingHelpers[info];
            }
            else
            {
                BuildHelpers();
            }
            return myMappingHelpers[info];
        }

        private void BuildHelpers()
        {
            // If can not find helper , then generate helper.
            ArrayList list = new ArrayList();
            foreach (ORMTypeInfo item in this.Project.Types)
            {
                if (myMappingHelpers.ContainsKey(item) == false)
                {
                    list.Add(item);
                    //ORMTypeMappingHelper helper = new ReflectionORMTypeMappingHelper(item);
                    //helper.Framework = this;
                    //myMappingHelpers[item] = helper;
                }
            }
            if (list.Count > 0)
            {
                XDesigner.Script.DynamicCompiler compiler = new XDesigner.Script.DynamicCompiler();
                compiler.Language = XDesigner.Script.CompilerLanguage.CSharp;
                compiler.ReferenceAssemblies.Add("System.dll");
                compiler.ReferenceAssemblies.Add("System.Data.dll");
                compiler.ReferenceAssemblies.Add("System.Xml.dll");

                List<Type> referencesType = new List<Type>();
                string strNameSpace = "NS" + Guid.NewGuid().ToString("N");

                ORMContext context = new ORMContext(this);

                compiler.SourceCode = CodeGenerator.GenerateClassesCode(
                    context,
                    (ORMTypeInfo[])list.ToArray(typeof(ORMTypeInfo)),
                    strNameSpace,
                    referencesType);
                referencesType.Add(this.GetType());
                referencesType.Add(typeof(ORMEngine));
                foreach (Type type in referencesType)
                {
                    string location = type.Assembly.Location;
                    if (compiler.ReferenceAssemblies.Contains(location) == false)
                    {
                        compiler.ReferenceAssemblies.Add(location);
                    }
                }

                if (compiler.Compile())
                {
                    System.Reflection.Assembly asm = compiler.ResultAssembly;
                    foreach (Type t in asm.GetTypes())
                    {
                        if (t.Namespace == strNameSpace)
                        {
                            foreach (ORMTypeInfo ot in list)
                            {
                                string typeName = CodeGenerator.ToTypeNameCode(ot.InstanceType, false)
                                    + "ORMTypeMappingHelper";

                                if (t.Name == typeName)
                                {
                                    ORMTypeMappingHelper helper = (ORMTypeMappingHelper)
                                        System.Activator.CreateInstance(t);
                                    helper.Framework = this;
                                    helper.ORMType = ot;
                                    myMappingHelpers[ot] = helper;
                                    break;
                                }//if
                            }//foreach
                        }//if
                    }//foreach
                }
                else
                {
                    System.Diagnostics.Debug.WriteLine(
                        compiler.SourceCodeWithCompilerErrorMessage);
                    string msg = string.Format(
                        ORMStrings.DynamicCompilerError_Message,
                        compiler.CompilerErrorMessage);
                    if (this.Config.ThrowDynamicCompilerError)
                    {
                        throw new Exception(msg);
                    }
                }
            }//if
        }

        public void SetHelper(Type type, ORMTypeMappingHelper helper)
        {
            if (type == null)
            {
                throw new ArgumentNullException("type");
            }
            if (helper == null)
            {
                throw new ArgumentNullException("helper");
            }
            ORMTypeInfo info = this.Project.Types[type];
            if (info == null)
            {
                throw new NotSupportedException(type.FullName);
            }
            helper.ORMType = info;
            myMappingHelpers[info] = helper;
        }

        public ORMTypeMappingHelper GetHelper(Type type)
        {
            if (type == null)
            {
                throw new ArgumentNullException("type");
            }
            ORMTypeInfo ot = this.Project.Types[type];
            if (ot == null)
            {
                throw new NotSupportedException(type.FullName);
            }
            return GetHelper(ot);
        }

        private ORMInstanceBuffer _Buffer = new ORMInstanceBuffer();

        public ORMInstanceBuffer Buffer
        {
            get { return _Buffer; }
        }

        private ORMConfig _Config = new ORMConfig();

        public ORMConfig Config
        {
            get
            {
                if (_Config == null)
                {
                    _Config = new ORMConfig();
                }
                return _Config; 
            }
            set
            {
                _Config = value; 
            }
        }

        //private ORMBufferLevel _BufferLevel = ORMBufferLevel.Specify;

        //public ORMBufferLevel BufferLevel
        //{
        //    get { return _BufferLevel; }
        //    set { _BufferLevel = value; }
        //}

        private ORMEventSender _EventSender = new ORMEventSender();
        /// <summary>
        /// ORM event sender
        /// </summary>
        public ORMEventSender EventSender
        {
            get { return _EventSender; }
            set { _EventSender = value; }
        }

        public ORMTypeInfo AddType(Type instanceType)
        {
            ORMTypeInfo info = this.Project.AddType(instanceType);
            return info;
        }

        private ORMDataBase _DataBase = new ORMDataBase();
        /// <summary>
        /// Data base handler
        /// </summary>
        public ORMDataBase DataBase
        {
            get { return _DataBase; }
            set { _DataBase = value; }
        }


        public event ORMEventHandler CloseConnectionEvent = null;

        public virtual void CloseConnection()
        {
            this.DataBase.Close();
            if (CloseConnectionEvent != null)
            {
                ORMEventArgs args = new ORMEventArgs( this , ORMEventStyle.CloseConnection ); 
                CloseConnectionEvent(this, args );
            }
        }

        public event ORMEventHandler OpenConnectionEvent = null;

        public virtual bool OpenConnection()
        {
            if (OpenConnectionEvent != null)
            {
                ORMEventArgs args = new ORMEventArgs( this , ORMEventStyle.OpenConnection );
                OpenConnectionEvent(this, args );
                if( args.Cancel )
                {
                    return false ;
                }
                if (args.Result is System.Data.IDbConnection)
                {
                    this.DataBase.Connection = ( IDbConnection ) args.Result;
                }
            }
            this.DataBase.Open();
            return true;
        }

        private void CheckConnection()
        {
            if (this.DataBase.IsOpen == false)
            {
                OpenConnection();
                if (this.DataBase.IsOpen == false )
                {
                    throw new InvalidOperationException("No connection");
                }
            }
        }

        public IDbCommand CreateDBCommand()
        {
            CheckConnection();
            return this.DataBase.CreateCommand();
        }

        //private ParameterStyle _ParameterStyle = ParameterStyle.SQLServerStyle;
        ///// <summary>
        ///// Parameter style in sql command text
        ///// </summary>
        //[System.ComponentModel.DefaultValue( ParameterStyle.SQLServerStyle )]
        //public ParameterStyle ParameterStyle
        //{
        //    get 
        //    {
        //        return _ParameterStyle; 
        //    }
        //    set
        //    {
        //        _ParameterStyle = value; 
        //    }
        //}

        //private bool _DebugMode = false;
        ///// <summary>
        ///// Whether framework in debug mode,=
        ///// </summary>
        //[System.ComponentModel.DefaultValue( false )]
        //public bool DebugMode
        //{
        //    get { return _DebugMode; }
        //    set { _DebugMode = value; }
        //}

        private Dictionary<string, int> myIncreaseIDs
            = new Dictionary<string, int>();

        public int GenerateIncreaseID(ORMContext context , string tableName , string keyFieldName )
        {
            if (tableName == null 
                || tableName.Trim().Length == 0)
            {
                throw new ArgumentNullException("name");
            }
            lock (myIncreaseIDs)
            {
                tableName = tableName.Trim().ToLower();
                if (myIncreaseIDs.ContainsKey(tableName))
                {
                    int v = myIncreaseIDs[tableName];
                    v++;
                    myIncreaseIDs[tableName] = v;
                    return v;
                }
                else
                {
                    string strSQL = "Select Max( "
                                    + keyFieldName + " ) + 1 From "
                                    + tableName;
                    context.Command.CommandText = strSQL;
                    context.Command.Parameters.Clear();
                    this.BeforeExecuteCommand(context, context.Command);

                    int intIncreaseID = 1;

                    object id = context.Command.ExecuteScalar();
                    if (id != null && DBNull.Value.Equals(id) == false)
                    {
                        intIncreaseID = Convert.ToInt32(id);
                    }
                    myIncreaseIDs[tableName] = intIncreaseID;
                    return intIncreaseID;
                }
            }//lock
        }

        public ORMBindingSource CreateBindingSource( Type instanceType , IEnumerable instances)
        {
            ORMBindingSource bs = new ORMBindingSource();
            bs.Framework = this;
            bs.ORMType = this.Project.Types[instanceType];
            foreach (object obj in instances)
            {
                bs.Add(obj);
            }
            return bs;
        }

        public virtual object CreateInstance(ORMTypeInfo type )
        {
            if (type == null)
            {
                throw new ArgumentNullException("type");
            }
            if (type.InstanceType == null)
            {
                throw new InvalidOperationException("type.InstanceType is null");
            }
            return System.Activator.CreateInstance(type.InstanceType);
        }

        public virtual void BeforeExecuteCommand(ORMContext context , ORMDBCommand command)
        {
            if (this.Config.DebugMode)
            {
                command.DebugPrint(this.Name);
            }
        }

        #region Read instance from data base 

        public Array ReadInstances(QueryConditionList conditions, Type instanceType)
        {
            if (instanceType == null)
            {
                throw new ArgumentNullException("instanceType");
            }
            ORMTypeInfo ot = this.Project.Types[instanceType];
            if (ot == null)
            {
                throw new NotSupportedException(instanceType.FullName);
            }
            ORMTypeMappingHelper helper = this.GetHelper(ot);
            if (helper == null)
            {
                throw new NotSupportedException(instanceType.FullName + " Helper");
            }
            using (ORMDBCommand cmd = new ORMDBCommand(this))
            {
                string strSQL = helper.SelectCommandText;
                string strWhere = ORMDBCommand.BuildConditionSQL(
                    conditions,
                    this.Config.ParameterStyle,
                    cmd);
                if (strWhere != null && strWhere.Length > 0)
                {
                    strSQL = "\r\n Where " + strWhere;
                }
                cmd.CommandText = strSQL;
                return ReadInstances(cmd, instanceType);
            }
        }

        public Array ReadInstances(string commandText, Type instanceType)
        {
            if (Util.HasContent(commandText))
            {
                using (ORMDBCommand cmd = new ORMDBCommand(this))
                {
                    cmd.CommandText = commandText;
                    return ReadInstances(cmd, instanceType);
                }
            }
            else
            {
                return null;
            }
        }

        public QueryConditionList CreateConditions(
            ORMContext context ,
            object instance, 
            IEnumerable conditionPropertyNames)
        {
            if (instance == null)
            {
                throw new ArgumentNullException("instance");
            }

            ORMTypeInfo ot = this.Project.Types[instance.GetType()];
            if (ot == null)
            {
                throw new NotSupportedException(instance.GetType().FullName);
            }
            ORMTypeMappingHelper helper = this.GetHelper( ot);
            QueryConditionList list = new QueryConditionList();
            foreach (ORMPropertyInfo property in ot.Properties)
            {
                foreach (string name in conditionPropertyNames)
                {
                    if (string.Compare(name, property.Name, true) == 0)
                    {
                        QueryCondition item = new QueryCondition();
                        item.FieldName = property.RuntimeDBFieldName;
                        item.Name = property.Name;
                        item.Value = helper.GetDBPropertyValue( context , instance, property.Name);
                        list.Add(item);
                        break;
                    }//if
                }//foreach
            }//foreach

            return list;
        }

        public Array ReadInstancesByCondition(
            object conditionInstance,
            System.Collections.IEnumerable conditionPropertyNames)
        {
            if (conditionInstance == null)
            {
                throw new ArgumentNullException("conditionInstance");
            }
            if (conditionPropertyNames == null)
            {
                throw new ArgumentNullException("conditionPropertyNames");
            }
            ORMTypeInfo ot = this.Project.Types[conditionInstance.GetType()];
            if (ot == null)
            {
                throw new NotSupportedException(conditionInstance.GetType().FullName);
            }
            ORMTypeMappingHelper helper = this.GetHelper(ot);
            if (helper == null)
            {
                throw new NotSupportedException( conditionInstance.GetType().FullName + " Helper" );
            }
            using (ORMDBCommand cmd = new ORMDBCommand(this))
            {
                string strSQL = helper.SelectCommandText;
                ORMContext context =  new ORMContext(this);
                context.ParameterStyle = this.Config.ParameterStyle;
                QueryConditionList list = CreateConditions(
                    context ,
                    conditionInstance,
                    conditionPropertyNames );
                if (list != null && list.Count > 0)
                {
                    string strWhere = ORMDBCommand.BuildConditionSQL(
                        list,
                        this.Config.ParameterStyle, 
                        cmd);
                    if (strWhere != null && strWhere.Length > 0)
                    {
                        strSQL = strSQL + "\r\nWhere " + strWhere;
                    }
                }
                cmd.CommandText = strSQL;
                return ReadInstances(cmd, conditionInstance.GetType() );
            }
        }

        //public Array ReadInstancesByCondition(string commandText , Type instanceType)
        //{
        //    if (instanceType == null)
        //    {
        //        throw new ArgumentNullException("instanceType");
        //    }
        //    ORMTypeInfo ot = this.Project.Types[instanceType];
        //    if (ot == null)
        //    {
        //        throw new InvalidOperationException("ORMTypeInfo is null");
        //    }
        //    ORMTypeMappingHelper helper = this.GetHelper(ot);
        //    if (helper == null)
        //    {
        //        throw new NotSupportedException(instanceType.FullName);
        //    }
        //    using (ORMDBCommand cmd = new ORMDBCommand(this))
        //    {
        //        string strSQL = helper.SelectCommandText;
        //        if (Util.HasContent(condition))
        //        {
        //            strSQL = strSQL + " Where " + condition;
        //        }
        //        cmd.CommandText = strSQL;
        //        return ReadInstances(cmd, instanceType);
        //    }
        //}

        /// <summary>
        /// Detect a instance exist in database.
        /// </summary>
        /// <param name="instanceType"></param>
        /// <param name="keyValue"></param>
        /// <returns></returns>
        public bool ExistInstance(Type instanceType, object keyValue)
        {
            if (instanceType == null)
            {
                throw new ArgumentNullException("instanceType");
            }
            ORMTypeInfo ot = this.Project.Types[instanceType];
            if (ot == null)
            {
                throw new InvalidOperationException("ORMTypeInfo is null");
            }
            if (ot.KeyProperty == null)
            {
                throw new InvalidOperationException("KeyProperty is null");
            }

            using( ORMDBCommand cmd = new ORMDBCommand( this ))
            {
                ORMContext context = new ORMContext( this );
                context.Command = cmd;
                ORMTypeMappingHelper helper = this.GetHelper(ot);
                StringBuilder mySQL = new StringBuilder();
                mySQL.Append(
                    "Select null from " + ot.TableName 
                    + " Where "
                    + ot.KeyProperty.RuntimeDBFieldName + " = ");
                ORMDBCommand.AppendParameterSQL(
                    mySQL,
                    ot.KeyProperty.RuntimeDBFieldName,
                    keyValue,
                    this.Config.ParameterStyle,
                    cmd);
                IDataReader reader = cmd.ExecuteReader(
                    CommandBehavior.SingleResult
                    | CommandBehavior.SingleRow);
                if (cmd.UserCancel)
                {
                    return false;
                }
                else
                {
                    bool result = reader.Read();
                    reader.Close();
                    return result;
                }
            }//using
        }

        public object ReadInstance(Type instanceType, object keyValue)
        {
            if (instanceType == null)
            {
                throw new ArgumentNullException("instanceType");
            }
            ORMTypeInfo ot = this.Project.Types[instanceType];
            if (ot == null)
            {
                throw new InvalidOperationException("ORMTypeInfo is null");
            }
            if (ot.KeyProperty == null)
            {
                throw new InvalidOperationException("KeyProperty is null");
            }
            object result = this.Buffer.GetBufferedInstance(keyValue, instanceType);
            if (result != null)
            {
                return result;
            }
            using ( ORMDBCommand cmd = new ORMDBCommand( this ))
            {
                ORMContext context = new ORMContext( this );
                context.Command = cmd;
                SetCurrentContext(context);
                ORMTypeMappingHelper helper = this.GetHelper(ot);
                StringBuilder mySQL = new StringBuilder();
                mySQL.Append(
                    helper.SelectCommandText 
                    + " Where " 
                    + ot.KeyProperty.RuntimeDBFieldName + " = ");
                ORMDBCommand.AppendParameterSQL(
                    mySQL, 
                    ot.KeyProperty.RuntimeDBFieldName, 
                    keyValue,
                    this.Config.ParameterStyle ,
                    cmd );
                context.Command.CommandText = mySQL.ToString();
                this.BeforeExecuteCommand(context, context.Command);

                List<RelationPropertyValue> relations = new List<RelationPropertyValue>();
                IDataReader reader = cmd.ExecuteReader(
                    CommandBehavior.SingleResult
                    | CommandBehavior.SingleRow);

                if (cmd.UserCancel)
                {
                    SetCurrentContext(null);
                    return null;
                }
                else
                {
                    Array array = InnerReadInstances(context, reader, ot, relations, 1);
                    reader.Close();
                    if (array != null && array.Length > 0)
                    {
                        result = array.GetValue(0);
                        if (relations != null && relations.Count > 0)
                        {
                            ReadRelationInstance(context, relations);
                        }
                    }
                    SetCurrentContext(null);
                    return result;
                }
            }//using
        }

        public System.Array ReadAllInstances(Type instanceType)
        {
            if (instanceType == null)
            {
                throw new ArgumentNullException("instanceType");
            }
            ORMTypeInfo ot = this.Project.Types[ instanceType ] ;

            using (ORMDBCommand cmd = new ORMDBCommand(this))
            {
                cmd.CommandText = this.GetHelper(ot).SelectCommandText;
                return ReadInstances(cmd, instanceType);
            }
        }

        public System.Array ReadInstances(ORMDBCommand cmd, Type instanceType)
        {
            if (instanceType == null)
            {
                throw new ArgumentNullException("instanceType");
            }
            if (cmd == null)
            {
                throw new ArgumentNullException("cmd");
            }
            //CurrentCommand = dbcmd;
            ORMContext context = new ORMContext(this);
            context.Command = cmd;
            SetCurrentContext(context);
            this.BeforeExecuteCommand(context, context.Command);
            Array result = InnerReadInstances(context, instanceType);
            //CurrentCommand = null;
            SetCurrentContext(null);
            return result;
        }

        private System.Array InnerReadInstances( ORMContext context , Type instanceType)
        {
            if ( context == null )
            {
                throw new ArgumentNullException("context");
            }
            if (instanceType == null)
            {
                throw new ArgumentNullException("instanceType");
            }
            ORMTypeInfo ot = this.Project.Types[instanceType];
            if (ot == null)
            {
                throw new NotSupportedException(instanceType.FullName);
            }
            
            System.Collections.ArrayList list = new System.Collections.ArrayList();
            System.Data.IDataReader reader = context.Command.ExecuteReader(
                System.Data.CommandBehavior.SingleResult);

            if (context.Command.UserCancel)
            {
                return null;
            }
            else
            {
                List<RelationPropertyValue> relationValues = new List<RelationPropertyValue>();

                Array result = InnerReadInstances(context, reader, ot, relationValues, 0);
                reader.Close();
                if (relationValues.Count > 0)
                {
                    ReadRelationInstance(context, relationValues);
                }
                return result;
            }
        }

        /// <summary>
        /// Read instances without read relation sub instances
        /// </summary>
        /// <param name="reader">database data reader</param>
        /// <param name="type"></param>
        /// <param name="relationValues"></param>
        /// <param name="maxRecordCount"></param>
        /// <returns></returns>
        private System.Array InnerReadInstances(
            ORMContext context ,
            IDataReader reader,
            ORMTypeInfo type,
            List<RelationPropertyValue> relationValues ,
            int maxRecordCount )
        {
            ORMPropertyInfo keyPropert = type.KeyProperty;
            ArrayList list = new ArrayList();
            ORMTypeMappingHelper helper = this.GetHelper(type);
            while (reader.Read())
            {
                if (maxRecordCount > 0 && list.Count >= maxRecordCount)
                {
                    break;
                }
                object instance = null;
                if (keyPropert != null)
                {
                    object keyValue = helper.GetDBValue(reader, keyPropert.Name);
                    if (keyValue != null && DBNull.Value.Equals(keyValue) == false)
                    {
                        if (this.Config.BufferLevel != ORMBufferLevel.None)
                        {
                            if ( this.Config.BufferLevel == ORMBufferLevel.Both || type.Buffered)
                            {
                                // get instance from global buffer
                                instance = this.Buffer.GetBufferedInstance(
                                    keyValue,
                                    helper.InstanceType );
                                if (instance != null)
                                {
                                    list.Add(instance);
                                    continue;
                                }
                            }
                        }
                        // get instance from context buffer
                        instance = context.Buffer.GetBufferedInstance(keyValue, helper.InstanceType);
                        if (instance != null)
                        {
                            list.Add(instance);
                            continue;
                        }
                    }
                }
                // create new instance
                instance = helper.CreateInstance();
                helper.FillInstance(context , reader, instance, relationValues);
                list.Add(instance);
                ORMPerformanceCounter.Increment(ORMPerformanceCounter.Name_Instance);
                if ( keyPropert != null )
                {
                    object keyValue = helper.GetPropertyValue(instance, keyPropert.Name);// keyPropert.GetValue(instance);
                    if (keyValue != null && DBNull.Value.Equals(keyValue) == false)
                    {
                        if (this.Config.BufferLevel != ORMBufferLevel.None)
                        {
                            if ( this.Config.BufferLevel == ORMBufferLevel.Both
                                || type.Buffered )
                            {
                                // save instance to global buffer
                                this.Buffer.SaveBufferedInstance(
                                    helper.InstanceType , 
                                    keyValue,
                                    instance);
                            }
                        }
                        // save instance to context buffer
                        context.Buffer.SaveBufferedInstance(
                            helper.InstanceType ,
                            keyValue,
                            instance);
                    }
                }
                // read next record
            }//while
            return list.ToArray(type.InstanceType);
        }
    
        private int ReadRelationInstance(
            ORMContext context ,
            List<RelationPropertyValue> relationValues )
        {
            int resultCount = 0;

            foreach (RelationPropertyValue item in relationValues)
            {
                ORMRelationStyle rs = item.Property.Relation.RuntimeStyle;
                if (rs == ORMRelationStyle.OneToMany
                    || rs == ORMRelationStyle.OneToManyWithKeyList)
                {
                    // get orm type of collection element
                    ORMTypeInfo type = this.Project.Types[
                        CollectionTypeHelper.GetElementType(item.Property.PropertyType)];
                    // create collection instance
                    object resultList = System.Activator.CreateInstance(
                        item.Property.PropertyType);
                    // get Add method info of collection
                    MethodInfo addMethod = CollectionTypeHelper.GetAddMethod(
                        item.Property.PropertyType);

                    ORMPropertyInfo childProperty = item.Property.Relation.ChildProperty;

                    ORMPropertyInfo keyProperty = type.KeyProperty;
                    if (childProperty == null)
                    {
                        childProperty = type.KeyProperty;
                    }

                    IDataReader reader = null;
                    if (rs == ORMRelationStyle.OneToMany)
                    {
                        ORMTypeMappingHelper helper = this.GetHelper(item.Property.OwnerType);
                        if (helper.FillDBCommand(
                            context,
                            helper.GetPropertyCommandText(item.Property.Name),
                            item.Instance,
                            item.Property.Name,
                            item.Value))
                        {
                            this.BeforeExecuteCommand(context, context.Command);

                            reader = context.Command.ExecuteReader(CommandBehavior.SingleResult);
                        }
                    }
                    else if (rs == ORMRelationStyle.OneToManyWithKeyList)
                    {
                        ORMTypeMappingHelper helper = this.GetHelper(item.Property.OwnerType);
                        if (Util.HasContent(item.Property.CommandText))
                        {
                            helper.FillDBCommand(
                                context,
                                helper.GetPropertyCommandText(item.Property.Name),
                                item.Instance,
                                item.Property.Name,
                                item.Value);
                        }
                        else if (keyProperty != null)
                        {
                            string strSQL = null;
                            strSQL = helper.SelectCommandText
                                + " Where " + keyProperty.RuntimeDBFieldName
                                + " In ( " + item.Value + " )";
                            context.Command.CommandText = strSQL;
                            context.Command.Parameters.Clear();
                        }

                        this.BeforeExecuteCommand(context, context.Command);

                        reader = context.Command.ExecuteReader(CommandBehavior.SingleResult);
                    }
                    if (reader != null)
                    {
                        List<RelationPropertyValue> subRelations
                            = new List<RelationPropertyValue>();
                        Array list = InnerReadInstances(
                            context,
                            reader,
                            type,
                            subRelations,
                            0);
                        reader.Close();
                        foreach (object obj in list)
                        {
                            addMethod.Invoke(resultList, new object[] { obj });
                        }

                        this.GetHelper(item.Property.OwnerType).SetPropertyValue(
                            item.Instance,
                            item.Property.Name,
                            resultList);

                        resultCount++;
                        if (subRelations.Count > 0)
                        {
                            resultCount += ReadRelationInstance(context, subRelations);
                        }
                    }
                }
                else if (rs == ORMRelationStyle.OneToOne)
                {
                    List<RelationPropertyValue> subRelations
                                    = new List<RelationPropertyValue>();
                    object propertyValue = null;
                    ORMTypeInfo type = item.Property.OwnerType;
                    ORMTypeMappingHelper helper = this.GetHelper(type);
                    if (Util.HasContent(item.Property.CommandText) == false)
                    {
                        // read instance from buffer
                        propertyValue = this.Buffer.GetBufferedInstance(item.Value, item.Property.PropertyType);
                    }
                    if (propertyValue == null)
                    {
                        if (helper.FillDBCommand(
                            context,
                            helper.GetPropertyCommandText(item.Property.Name),
                            item.Instance,
                            item.Property.Name,
                            item.Value))
                        {
                            this.BeforeExecuteCommand(context, context.Command);

                            IDataReader reader = context.Command.ExecuteReader(
                                CommandBehavior.SingleResult
                                | CommandBehavior.SingleRow);

                            if (reader != null)
                            {
                                ORMTypeInfo valueType = this.Project.Types[item.Property.PropertyType];

                                Array list = InnerReadInstances(
                                    context,
                                    reader,
                                    valueType,
                                    subRelations,
                                    1);
                                reader.Close();
                                propertyValue = list.GetValue(0);
                            }
                        }
                    }
                    if (propertyValue != null)
                    {
                        this.GetHelper(item.Property.OwnerType).SetPropertyValue(
                           item.Instance,
                           item.Property.Name,
                           propertyValue);

                        //this.GetHelper( item.Property.OwnerType ).SetPropertyValue(
                        //    item.Instance ,
                        //    item.Property.Name ,
                        //    instance );

                        resultCount++;
                        if (subRelations.Count > 0)
                        {
                            resultCount += ReadRelationInstance(context, subRelations);
                        }
                    }//if
                }
            }//foreach

            return resultCount ;
        }

        #endregion


        /// <summary>
        /// Update instance to database
        /// </summary>
        /// <param name="instance"></param>
        /// <returns></returns>
        public int Update(object instance)
        {
            if (instance == null)
            {
                throw new ArgumentNullException("instance");
            }
            ORMTypeInfo type = this.Project.Types[instance.GetType()];
            if (type == null)
            {
                throw new NotSupportedException(instance.GetType().FullName);
            }
            using ( ORMDBCommand cmd = new ORMDBCommand( this ))
            {
                ORMContext context = new ORMContext(this);
                context.Command = cmd;
                SetCurrentContext(context);
                ORMTypeMappingHelper helper = this.GetHelper(type);
                int result = 0;
                if (helper.PrepareUpdateCommand(context, instance))
                {
                    this.BeforeExecuteCommand(context, context.Command);
                    result = context.Command.ExecuteNonQuery();
                }
                SetCurrentContext(null);
                return result;
            }
        }

        /// <summary>
        /// Insert a instance into database
        /// </summary>
        /// <param name="instance"></param>
        /// <returns></returns>
        public int Insert(object instance)
        {
            if (instance == null)
            {
                throw new ArgumentNullException("instance");
            }
            ORMTypeInfo type = this.Project.Types[instance.GetType()];
            if (type == null)
            {
                throw new NotSupportedException(instance.GetType().FullName);
            }
            using ( ORMDBCommand cmd = new ORMDBCommand( this ))
            {
                ORMContext context = new ORMContext(this);
                context.Command = cmd;
                SetCurrentContext(context);
                ORMTypeMappingHelper helper = this.GetHelper(type);
                int result = 0;
                if (helper.PrepareInsertCommand(context, instance))
                {
                    this.BeforeExecuteCommand(context, context.Command);
                    result = context.Command.ExecuteNonQuery();
                }
                SetCurrentContext(null);
                return result;
            }
        }

        public virtual bool GenerateInstaceID(object instance)
        {
            if (instance == null)
            {
                throw new ArgumentNullException("instance");
            }

            ORMTypeInfo type = this.Project.Types[instance.GetType()];
            object id = GenerateID( instance.GetType() );
            ORMPropertyInfo keyProperty = type.KeyProperty;
            ORMTypeMappingHelper helper = this.GetHelper(type);
            helper.SetPropertyValue(instance, keyProperty.Name, id);
            return true;
        }

        public virtual object GenerateID(Type instanceType)
        {
            if (instanceType == null)
            {
                throw new ArgumentNullException("instanceType");
            }

            ORMTypeInfo type = this.Project.Types[instanceType];
            ORMTypeMappingHelper helper = this.GetHelper(type);
            if (helper == null)
            {
                return null;
            }
            else
            {
                using ( ORMDBCommand cmd = new ORMDBCommand( this ))
                {
                    ORMContext context = new ORMContext(this);
                    context.Command = cmd;
                    SetCurrentContext(context);
                    object result = helper.GenerateID( context );
                    SetCurrentContext(null);
                    return result;
                }
            }
        }

        #region Delete instance from data base

        /// <summary>
        /// Delete a instance 
        /// </summary>
        /// <param name="instance">instance</param>
        /// <returns>number of data base records has been deleted.</returns>
        public int Delete(object instance)
        {
            if( instance == null )
            {
                throw new ArgumentNullException("instance");
            }
            ORMTypeInfo type = this.Project.Types[ instance.GetType()] ;
            if( type == null )
            {
                throw new NotSupportedException( instance.GetType().FullName );
            }
            using ( ORMDBCommand cmd = new ORMDBCommand( this ))
            {
                ORMContext context = new ORMContext( this );
                context.Command = cmd;
                SetCurrentContext(context);
                int result = InnerDelete(context, instance);
                SetCurrentContext(null);
                return result;
            }
        }

        private int InnerDelete(ORMContext context, object instance)
        {
            ORMTypeInfo type = this.Project.Types[instance.GetType()];
            if (type == null)
            {
                throw new NotSupportedException(instance.GetType().FullName);
            }
            ORMTypeMappingHelper helper = this.GetHelper(type);
            int result = 0;
            if (helper.PrepareDeleteCommand(context, instance))
            {
                this.BeforeExecuteCommand(context, context.Command);
                result = context.Command.ExecuteNonQuery();
            }
            if (result > 0)
            {
                this.Buffer.RemoveBufferedInstance(helper.GetKeyValue(instance), instance);
                foreach (ORMPropertyInfo property in type.Properties)
                {
                    if (property.Ignore || property.Relation == null)
                    {
                        continue;
                    }
                    if (property.Relation.UniteDelete == false)
                    {
                        continue;
                    }
                    // delete relation instance
                    ORMRelationStyle rs = property.Relation.RuntimeStyle ;
                    if ( rs == ORMRelationStyle.OneToOne)
                    {
                        object obj = helper.GetPropertyValue(instance, property.Name);
                        if (obj != null)
                        {
                            result += InnerDelete(context, obj );
                        }
                    }
                    else if (rs == ORMRelationStyle.OneToMany
                        || rs == ORMRelationStyle.OneToManyWithKeyList)
                    {
                        object list = helper.GetPropertyValue(instance, property.Name);
                        if (list != null)
                        {
                            if (list is System.Collections.IEnumerable)
                            {
                                foreach (object obj in (System.Collections.IEnumerable)list)
                                {
                                    result += InnerDelete(context, obj);
                                }
                            }
                        }
                    }//end if
                }//foreach
            }//if
            return result;
        }

        #endregion

        public void Dispose()
        {
            this._DataBase.Dispose();
            if (System.Threading.Thread.GetData(myDataStoreSlot) == this)
            {
                System.Threading.Thread.SetData(myDataStoreSlot, null);
            }
            this._EventSender.Clear();
            this._Buffer.Clear();
        }
    }
}
